/*
Copyright (c) 2009, Joel Reardon
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the University of Waterloo nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef __SPE_H_
#define __SPE_H_

#include <glib.h>
#include <pthread.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#define SPE 0x112233AA
#define SPE_VARSTATE 0x11223344
#define SPE_VARSTATE_DATA 0x11223355

#define SPE_VARFAMILY 0x33445566
#define SPE_VAR 0x66442200

#define SPE_DATA_ARRAY 0x66554433
#define SPE_DATA_LIST 0x22334455

#define SPE_SOCKET_ERROR_BIND 5
#define SPE_SOCKET_ERROR_SOCKET 4
#define SPE_SOCKET_SUCCESS 0

#define SPE_OBSERVER_BEGIN -1
#define SPE_OBSERVER_OBSERVE 0
#define SPE_OBSERVER_END 1




/*
typedef struct {
	GQueue * Q; 
	int account;
	int cur_size;
	struct timeval last_checked;
	struct timeval start_timing;
	pthread_mutex_t lock;
	void * buf_id;
	int enabled;
} spe_buffer_t;



struct spe_buffer_manager {
	GHashTable* data;
};

typedef struct spe_buffer_manager spe_buffer_manager_t;


typedef struct {
	struct timeval time;
	int len;
} spe_buffer_time_t;

*/

typedef struct {
	FILE* f;
	pthread_mutex_t lock;
	int enabled;	
} spe_logger_t;


typedef struct {
	int magic;
	char* name;
	GHashTable* data;
	int data_type;
	int dataset_size;
} spe_varfamily_t;


typedef struct {
	int magic;
	struct timeval timing;
	int is_valid;
	pthread_mutex_t lock;
	void* data;
} spe_var_t;


struct spe_data_list {
	int magic;
	GSList *results;
};

typedef struct spe_data_list spe_data_list_t;
	
struct spe_data_array {
	int magic;
	int *results;
	int curpos;
 	int size;
	int dataset_size;
};

typedef struct spe_data_array spe_data_array_t;


typedef struct {
	int magic;
	struct timeval timing;
	int cur_state;
	int dataset_size;
	pthread_mutex_t lock;
	GHashTable* states;
} spe_varstate_t;


struct spe_varstate_data{
	int magic;
	int count;
	int state;
	int dataset_size;
	void* data;

};

typedef struct spe_varstate_data spe_varstate_data_t;



typedef struct {
	void* observee;
	char* name;
	int family;
	spe_logger_t * logger;
	char enabled;
	void* state;
	void (*state_delete) (void*);
	int quota;
	int * metric;
	struct timeval next_time;	
	struct timeval interval;
	pthread_mutex_t lock;
	int interest;
} spe_observer_t;

//typedef struct spe_observer spe_observer_t;
typedef void* spe_obs_func;

struct spe_fl { 
	int* symbol_link;
	int* symbol_state;
	pthread_mutex_t *lock;
	pthread_t unloader;
	void * dlhandle;
	FILE * fd;
	
};

typedef struct spe_fl spe_fl_t;


struct spe_observer_manager {
	spe_fl_t * fl;
	GHashTable* data;
	
	GQueue *Q;
	pthread_t Qthread;
	pthread_cond_t *Qcond;
	pthread_mutex_t *Qlock;
	int Qthread_quit;
	spe_logger_t * logger;
};

typedef struct spe_observer_manager spe_observer_manager_t;

//TODO add the ability to control logic with generalized toggles

/// spe_t is the main data type for spe. Provides the context for all 
/// operations that are done in the system. Generally only one is needed
/// for a single application, and _spe is defined for that purpose.
typedef struct {
	int magic;
	struct timeval init_time;
	GHashTable* spe_families;
//	spe_buffer_manager_t * buf_mgr;
	spe_observer_manager_t* obs_mgr;
	
	int dataset_size;
	
	int socket_port;
	int socket_thread_keepalive;
	int socket;
	pthread_t socket_thread;
	pthread_mutex_t socket_lock;
	
	int alive;
	int collect_indexed_vars;
	

} spe_t;

extern spe_t* _spe;

typedef struct {
	int family;
	int newstate;
	int quota;
	struct timeval tv;
	char* filename;
	spe_t * spe;
} set_state_t; 

///utility methods
int spe_time_diff(struct timeval* later, struct timeval* earlier);
void spe_time_copy(struct timeval* dst, struct timeval* src);
void spe_time_add(struct timeval* dst, struct timeval* src);
void spe_gettimeofday(struct timeval* tv);
void spe_gettime(spe_var_t * var);
int spe_gettimediff(spe_var_t * var);
void spe_variable_stat(spe_t* spe, int val);

///spe methods
spe_t* spe_init(); //initializes a basic spe object

//initializes an spe object allowing the port and dataset size to be set
spe_t* spe_init_full(int port, int dataset_size); 

//if the basic initializer is used, the reinitialize method can be used to
//later set the dataset size and the port for interaction
void spe_reinit_full(spe_t* spe,int port, int dataset_size);

//frees the memory used by spe
void spe_delete(spe_t*);

//deletes the spe object and stops the interaction socket
void spe_delete_full(spe_t * context); 

//Maintainance thread no longer used
//void spe_maintainance_thread_start(spe_t*, int trim_size);
//void spe_maintainance_thread_stop(spe_t*);

//socket thread start and stops begin and end the interaction socket
//used for communication with spe outside the executable
void spe_socket_thread_start(spe_t*, int port);
void spe_socket_thread_stop(spe_t*);

//functions to control access to individual variables being managed
void spe_var_length(spe_var_t * var);
void spe_delete(spe_t * stat);
void spe_timing_start(spe_t * context, const char* str, void* p);
void* spe_var_fetch(spe_t * context, const gchar* str, void* p);
void spe_timing_stop(spe_t * context, const gchar* str, void* p);
void spe_timing_period(spe_t * context, const gchar* str, void* p);
void spe_data_add(spe_t * context, const gchar* str, void* p, int data);
int spe_state_change(spe_t * context, const gchar* str, void* p,int newstate);

//spe_varfam methods
spe_varfamily_t* spe_varfamily_new(const char* name, int data_type, int dataset_size);
void spe_varfamily_delete(spe_varfamily_t*);
spe_varfamily_t* spe_varfamily_add(spe_t * context, const char* str, int data_type, int dataset_size);
void spe_switch_var_delete(void* var);

//spe_varstate methods
spe_varstate_t* spe_varstate_new(int dataset_size);
void spe_varstate_delete(spe_varstate_t* varstate);
spe_varstate_t* spe_varstate_add(spe_varfamily_t* varfam, void* p);
spe_varstate_t * spe_varstate_fetch_add(spe_t * context, const char* str, void* p);


//spe_varstate_data methods
void spe_varstate_data_delete(spe_varstate_data_t * vsdata);
spe_varstate_data_t * spe_varstate_data_new(int state, int dataset_size);

//spe_var_data methods
void* spe_var_data_new(int dataset_size);
spe_data_list_t *spe_data_list_new();
spe_data_array_t *spe_data_array_new(int dataset_size);
void spe_var_data_delete(void* data);
void spe_var_data_sort(void * var, int len);

//spe_var methods
spe_var_t* spe_var_new(int dataset_size);
void spe_var_delete(spe_var_t *);
spe_var_t* spe_var_add(spe_varfamily_t* varfam, void* p);
spe_var_t * spe_var_fetch_add(spe_t * context, const char* str, void* p);
void spe_data_add_impl( void * var, int val);
int spe_var_data_length(void* var);

//buffer methods: no longer used
/*spe_buffer_manager_t * spe_buffer_manager_new();
void spe_buffer_manager_delete(spe_buffer_manager_t * bm);
spe_buffer_t * spe_buffer_new();
void spe_buffer_delete(spe_buffer_t * buf);
int spe_buffer_register ( spe_t *context, void* buf_id);
int spe_buffer_deregister( spe_t * context, spe_buffer_t* buf);
int spe_buffer_insert(spe_t *context, void *buf_id, int len, int cur_size);
int spe_buffer_remove(spe_t * context, void *buf_id, int len, int cur_size);
int spe_buffer_size_change(spe_t * context, spe_buffer_t * buf, int delta, int size, struct timeval* time);
*/

//statistics and output methods for variables
void spe_clear_data(spe_t *context, const gchar* str);
double spe_get_stdev(void * data, double avr);
double spe_get_average(void * data);
void spe_var_write_cdf(void * data, FILE* file);
double spe_get_stdev_array(int* data, int n, double avr);
double spe_get_average_array(int* data, int n);
void spe_var_write_cdf_array(int* data, int n, FILE* file);
double spe_get_stdev_list(GSList* list, double avr);
double spe_get_average_list(GSList * list);
void spe_var_write_cdf_list(GSList * cur, FILE* file);

//output methods
void spe_generate_report(spe_t * context, const gchar* str);
void spe_write_report_entry( gpointer key, gpointer value, gpointer file);
void spe_generate_report_file(spe_t * context, FILE* fout);



/*
Registers a new observer for libspe. observee is the object that 
is being observed, and will be passed into the callback to make observations. 
char* is the friendlyname, willbe used for the filename.
interval is the time between successive calls to the callback for automatic 
observation. Set this to 0 to disable automatic obersvation
*/
spe_observer_t * spe_observer_register(spe_t* spe, char* name, void* observee, int family, int* metric);
spe_observer_t* spe_observer_register_time(spe_t* spe, char* name, void* observee, int family, int* metric, struct timeval * interval);
spe_observer_t * spe_observer_register_full(spe_t* spe, char* name, void* observee, int family, int* metric, struct timeval* interval, void* state, void* state_delete);
void spe_observer_deregister(spe_t* spe, spe_observer_t* ticket);
void spe_observee_deregister(spe_t* spe, void* observee);



void spe_observer_delete(spe_observer_t* obs);

void spe_observee_disable(spe_t* spe, void * observee);
void spe_observee_enable(spe_t* spe, void * observee, char*) ;
void spe_observer_disable(spe_observer_manager_t* spe, spe_observer_t * ticket);
void spe_observer_enable(spe_observer_manager_t* spe, spe_observer_t * ticket, char*);
void spe_observer_disable_fe(void* key, void* value, void* ptr);

void spe_observee_make(spe_t* spe, void* observee, int obs_type);
void spe_observation_make(spe_observer_manager_t* om, spe_observer_t* ticket, int obs_type);
void spe_observer_next_observation_time(spe_observer_t* obs);


spe_observer_manager_t * spe_observer_manager_new();
void spe_observer_manager_delete(spe_observer_manager_t * om);




void spe_observer_list(spe_t * context, GString* gs);
void spe_observer_list_fe(void* key, void* data, void* parm);
void spe_observer_view(spe_t * spe, GString* gs, void* key);
void spe_observer_dump(spe_t *spe, FILE* fdump);
void spe_observer_set_state(spe_t * spe, void* ptr, int newstate, int quota, struct timeval *tv, char* filename);

void spe_observer_manager_queue_remove(spe_observer_manager_t * om , spe_observer_t * obs);
void spe_observer_manager_queue_add(spe_observer_manager_t * om , spe_observer_t * obs);
void * spe_observer_manager_queue_thread ( spe_observer_manager_t * om);
void spe_observer_manager_queue_observation_make ( spe_observer_manager_t* om) ;
struct timeval* spe_observer_manager_queue_first_time(GQueue * Q);
	


/* spe_function_libary */
spe_fl_t * spe_fl_new();
void spe_fl_delete(spe_fl_t* fl);
spe_obs_func spe_fl_get(spe_fl_t *fl, int family);
void* spe_fl_set(spe_fl_t *fl, int family);
void spe_fl_resym_all(spe_fl_t* fl);

void* spe_fl_unload(void* handle);
char* spe_fl_load(spe_fl_t * fl, char* library);



/* spe_socket function */
int strpre(char* str, char* prefix); //checks if prefix appears as prefix in str
void spe_listener_command_dump(spe_t * spe, char* cmd, GString* ret);
void spe_listener_command_list(spe_t * spe, char* cmd, GString* ret);
void spe_listener_command_view(spe_t * spe, char* cmd, GString* ret);
void spe_listener_command_seto(spe_t * spe,char* cmd, GString* ret);
void spe_listener_command_setb(spe_t * spe,char* cmd, GString* ret);
void spe_listener_command_setv(spe_t * spe,char* cmd, GString* ret);
void spe_listener_command_flib(spe_t* spe, char *cmd, GString* ret);
void * spe_listener_interaction_thread (void*  param);
void* spe_listener_thread(spe_t * spe);



/* spe main interaction functions */
//void spe_buffer_set_state(spe_t* spe, void* ptr, int newstate);
//void spe_buffer_dump(spe_t *spe, FILE* fdump);
//void spe_buffer_list(spe_t *spe, GString* gs);


void spe_variable_set_state(spe_t* spe, void* ptr, int newstate);
void spe_variable_dump(spe_t *spe, FILE* fdump);
void spe_variable_list(spe_t *spe, GString* gs);


/* spe logger functions */
void spe_logger_delete(spe_logger_t * logger);
spe_logger_t * spe_logger_new();
void spe_logger_log(spe_logger_t * logger, const char* message, ...);



#endif


